module Main where

import Linear.V
import Data.Proxy
import GHC.TypeNats
import Data.Foldable

import Math.Alspach

testAlspach :: KnownNat n => Proxy n -> IO ()
testAlspach prox = putStrLn $ case failed of
    Nothing -> "Polynomial has no usable monomials for |S| = " ++ card
    Just [] -> "Conjecture verified for |S| = " ++ card
    Just fails -> "Could not verify conjecture for |S| = " ++ card ++
      " in field[s] of size[s] " ++ show fails
  where card = show (reflectDim prox)
        failed = failedPrimes candidateTermsViaBP' prox

testEvenAlspach :: KnownNat n => Proxy n -> IO ()
testEvenAlspach prox = putStrLn $ case failed of
    Nothing -> "Some polynomial has no usable monomials for |S| = " ++ card
    Just [] -> "Conjecture verified for |S| = " ++ card
    Just fails -> "Could not verify conjecture for |S| = " ++ card ++
      " in Z/2pZ when n is one of " ++ show fails
  where card = show (reflectDim prox)
        failed = failedPrimesTwo prox

main :: IO ()
main = do
  let primeMax = 9
      twoMax = 9
      test1 n = reifyDimNat n testAlspach
      test1Two n = reifyDimNat n testEvenAlspach

  putStrLn $ "Checking Alspach for prime fields through |S| = " ++
    show primeMax ++ "..."
  for_ [1..primeMax] test1

  putStrLn $ "Checking Alspach for Z/2pZ through |S| = " ++
    show twoMax ++ "..."
  for_ [1..twoMax] test1Two

