{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE TypeOperators #-}
{-# LANGUAGE DataKinds #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE DeriveTraversable #-}
{-# LANGUAGE TupleSections #-}
module Math.AlspachProfile where

import Control.Monad
import Data.Proxy
import Control.Monad.Random.Class
import System.Random.Shuffle
import System.Microtimer
import Data.Traversable
import GHC.TypeNats
import qualified Data.Map.Strict as M

import Math.Polynomial
import Math.Alspach

-- TODO proper force
timeBP :: (KnownNat n, Num a, Eq a) =>
  Monomial n -> Expr n a -> IO Double
timeBP bound factors =
  time_ (return $! length (terms (boundedProductE bound factors)))


-- TODO this is not evenly distributed---but do I *want* it to be?
genGrouping :: MonadRandom m => a -> [a] -> m (Grouping a)
genGrouping z [] = return (Leaf z)
genGrouping _ [x] = return (Leaf x)
genGrouping z xs = do
  splitPoint <- getRandomR (1, length xs - 1)
  let (xs1, xs2) = splitAt splitPoint xs
  left <- genGrouping z xs1
  right <- genGrouping z xs2
  return (Prod left right)

genPermGrouping :: MonadRandom m => a -> [a] -> m (Grouping a)
genPermGrouping z = shuffleM >=> genGrouping z

samplePolyTimings :: KnownNat n => Proxy n -> Int ->
  IO [(Expr n Integer, Double)]
samplePolyTimings prox count = do
  groupings <- replicateM count (genPermGrouping 1 nAlspachInstPolyFactors)
  for groupings $ \g -> (g, ) <$> timeBP (mBound prox) g

sampleTimings :: KnownNat n => Proxy n -> Int -> IO [Double]
sampleTimings prox count = map snd <$> samplePolyTimings prox count


data MulInfo =
  MulInfo {
    termsL :: Int,
    termsR :: Int,
    afterGrouping :: Int,
    afterCanceling :: Int,
    afterBounding :: Int
  } deriving (Show, Read)

prodWithStats :: (KnownNat n, Num a, Eq a) =>
  Monomial n -> Polynomial n a -> Polynomial n a -> (Polynomial n a, MulInfo)
prodWithStats bound p q = (bounded, info)
  where grouped = M.unionsWith (+) $ [_getCoeffs (t *! p) | t <- terms q]
        canceled = fromCoeffs grouped
        bounded = filterTerms (\(T (m, _)) -> m <<= bound) canceled
        info = MulInfo (length p) (length q) (length grouped)
                       (length canceled) (length bounded)

data GMulInfo a = GMILeaf a | GMIProd MulInfo (GMulInfo a) (GMulInfo a)
  deriving (Show, Read, Functor, Foldable, Traversable)

loggingBoundedProductE :: forall n a. (KnownNat n, Num a, Eq a) =>
  Monomial n -> Expr n a -> GMulInfo (Polynomial n a)
loggingBoundedProductE bound = snd . foldMapG start step
  where start p = (p, GMILeaf p)
        step (p, l) (q, r) = (\i -> GMIProd i l r) <$> prodWithStats bound p q

